from .__version__ import __version__
from .myforestplot import *
from .cleaning_utils import (
    statsmodels_fitting_result_dataframe,
    add_pretty_risk_column,
    count_category_frequency,
    sort_category_item,
    statsmodels_pretty_result_dataframe,
)




from typing import Union, Optional, List, Dict, Tuple, Any
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec


def obtain_indexes_from_category_item(ser_cate : pd.Series,
                                      ser_item : pd.Series
                                      ) -> Tuple[np.array, np.array]:
    """Create index for category and item from series of category and item
    for vertically aligned labels and errorbar plot.
    It is noted that index has negative continuous values, starting from 0 to -n.

    Args:
        ser_cate: Series of categories matched with ser_item.
        ser_item: Series of items.

    Returns:
        Indexes for caategories and items.
    """
    ser_cate = ser_cate.copy()

    n_cate = len(ser_cate.unique())
    ser_cate[ser_cate.duplicated()] = np.nan
    y_index = []
    y_index_cate = []
    index = 0
    for cate, item in zip(ser_cate, ser_item):
        if cate != cate:
            y_index.append(index)
        else:
            y_index_cate.append(index)
            index -= 1
            y_index.append(index)
        index -= 1

    y_index_cate = np.array(y_index_cate)
    y_index = np.array(y_index)

    return(y_index_cate, y_index)


def errorbar_forestplot(
    ax: plt.Axes,
    y_index: np.array,
    df: Optional[pd.DataFrame] = None,
    risk: str = "risk",
    lower: Union[str, int] = 0,
    upper: Union[str, int] = 1,
    y_adj: float = 0,
    errorbar_kwds: Optional[dict] = None,
    ref_kwds: Optional[dict] = None,
    errorbar_color: Optional[str] = None,
    ref_color: Optional[str] = None,
    label: Optional[str] = None,
    log_scale: bool = False,
    ):
    """Error bar plot for a forest plot.

    Args:
        ax: Axis to be drawn.
        y_index: index to be plotted.
        risk: Column name for risk.
        lower: Column name for lower confidence interval.
        upper: Column name for upper confidence interval.
        y_adj: For this value, plotting is moved.
        errorbar_kwds: Passed to ax.errorbar function.
        ref_kwds: Passed to ax.scatter function.
        df: Dataframe for another result.
        label: Label for stratified drawings. Passed to ax.errorbar.
        log_scale: Plot risk in log scale (np.log).
    """
    y_index = y_index + y_adj

    df = df.copy()
    if errorbar_kwds is None:
        errorbar_kwds = dict(fmt="o",
                             capsize=5,
                             markeredgecolor="black",
                             ecolor="black",
                             color='white'
                             )
    if ref_kwds is None:
        ref_kwds = dict(marker="s", s=20, color="black")

    if log_scale:
        df[risk] = np.log(df[risk])
        df[lower] = np.log(df[lower])
        df[upper] = np.log(df[upper])

    df["xerr_lower"] = df[risk] - df[lower]
    df["xerr_upper"] = df[upper] - df[risk]

    cond = df[risk].notnull()
    ax.errorbar(df.loc[cond, risk],
                y_index[cond],
                xerr=df.loc[cond, ["xerr_lower", "xerr_upper"]].T,
                label=label,
                **errorbar_kwds
                )

    cond = df[risk].isnull()
    ref_v = 0 if log_scale else 1
    df["ref"] = df[risk].mask(cond, ref_v).mask(~cond, np.nan)
    ax.scatter(df["ref"], y_index, **ref_kwds)


def embed_strings_forestplot(
    ax: plt.Axes,
    ser: pd.Series,
    y_index: np.array,
    x: float,
    header: str,
    fontsize: int = None,
    y_header: float = 1.0,
    text_kwds: Optional[dict] = None,
    header_kwds: Optional[dict] = None,
    replace: Optional[dict] = None,
    ):
    """Embed strings/values of one column with header.

    Args:
        ser: Values of this series will be embedded.
        x: x axis value of text position, ranging from 0 to 1.
        df: Dataframe for another result.
    """
    if text_kwds is None:
        text_kwds = {}
    if header_kwds is None:
        header_kwds = {}
    ax.text(x, y_header, header, ha="left", va="center",
            fontsize=fontsize, **header_kwds)

    if replace is not None:
        ser = ser.replace(replace)

    for y, text in zip(y_index, ser):
        ax.text(x, y, text, ha="left", va="center",
                fontsize=fontsize, **text_kwds)




from typing import Union, Optional, List, Dict, Tuple, Any
from dataclasses import dataclass, field
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import statsmodels

import myforestplot.vis_utils as vis_utils


@dataclass(repr=True)
class BaseForestPlot():
    """Class for creating a forest plot.

    Args:
        ratio: Ratio for text part and figure part.
        df: Engineered dataframe.
        hide_spines : Hide outlines of axes.  Takes "right","top","bottom","left" or these list.
        vertical_align: Align categorical names above items. It requires dataframe to have 
            "category" and "item" column names.
    """
    df: pd.DataFrame
    ratio: Tuple[float, float] = (8,3)
    figsize: Tuple[float, float] = (5,3)
    dpi: int = 150
    hide_spines: List[str] = field(default_factory=lambda: ["left", "top", "right"]) 
    yticks_show: bool = False
    yticklabels_show: bool = False
    xticks_show: bool = True
    text_axis_off: bool = True
    vertical_align: bool = False

    def __post_init__(self):
        self.df = self.df.reset_index(drop=True)

        self.create_y_index()
        self.figure_layout()

    def create_y_index(self):
        self.n_item = self.df.shape[0]
        if self.vertical_align:
            if "category" not in self.df.columns:
                raise Exception("Need 'category' column for df variable.")
            if "item" not in self.df.columns:
                raise Exception("Need 'item' column for df variable.")

            self.y_index_cate, self.y_index = (
                vis_utils.obtain_indexes_from_category_item(
                    self.df["category"],
                    self.df["item"]
                    )
                )
        else:
            self.y_index = np.array([-i for i in range(self.n_item)])
            cond = self.df["category"].duplicated()
            self.y_index_cate = -np.array(cond[cond==False].index)

        self.ymax = np.max(self.y_index_cate)
        self.ymin = np.min(self.y_index)

    def figure_layout(self):
        """Create figure layout for plotting"""
        plt.rcParams["figure.facecolor"] = "white"
        plt.rcParams["savefig.facecolor"] = "white"

        self.fig = plt.figure(figsize=self.figsize, dpi=self.dpi)
        gs = GridSpec(1, sum(self.ratio))

        ss1 = gs.new_subplotspec((0, 0), colspan=self.ratio[0])
        ss2 = gs.new_subplotspec((0, self.ratio[0]), colspan=self.ratio[1])

        self.ax1 = plt.subplot(ss1)
        self.ax2 = plt.subplot(ss2, sharey=self.ax1)

        # Hide spines.
        if self.text_axis_off:
            self.ax1.set_axis_off()
        self.ax2.spines[self.hide_spines].set_visible(False)
        # Show or hide ticks
        if not self.xticks_show:
            self.ax2.set_xticks([])
        if self.yticks_show:
            self.ax2.set_yticks(self.y_index)
        else:
            self.ax2.set_yticks([])
        if not self.yticklabels_show:
            self.ax2.set_yticklabels([])

        # Figure restriction.
        self.ax2.set_ylim([self.ymin - 0.5, 0.5])

        # Create empty field for text part.
        self.ax1.set_xlim([0,1])
        self.ax1.scatter([0, 0, 1, 1], [self.ymin, 0, self.ymin, 0], color="white")

    def errorbar(self, 
                 risk: str = "risk", 
                 lower: Union[str, int] = 0, 
                 upper: Union[str, int] = 1, 
                 y_adj: float = 0,
                 errorbar_kwds: Optional[dict] = None,
                 ref_kwds: Optional[dict] = None,
                 df: Optional[pd.DataFrame] = None,
                 errorbar_color: Optional[str] = None,
                 ref_color: Optional[str] = None,
                 label: Optional[str] = None,
                 log_scale: bool = False,
                 ):
        """

        Args:
            risk: Column name for risk.
            lower: Column name for lower confidence interval.
            upper: Column name for upper confidence interval.
            y_adj: For this value, plotting is moved. 
            errorbar_kwds: Passed to ax.errorbar function.
            ref_kwds: Passed to ax.scatter function.
            df: Dataframe for another result.
            errorbar_color: If specified, ecolor and coloer in erorrbar_kwds is 
                changed to this value.
            ref_color: If specified, ecolor and coloer in ref_kwds is 
                changed to this value.
            label: Label for stratified drawings. Passed to ax.errorbar.
            log_scale: Plot risk in log scale (np.log).
        """
        if df is None:
            df = self.df
        if errorbar_color is not None:
            errorbar_kwds["ecolor"] = errorbar_color
            errorbar_kwds["color"] = errorbar_color
        if ref_color is not None:
            ref_kwds["color"] = ref_color

        vis_utils.errorbar_forestplot(
            ax=self.ax2, 
            y_index=self.y_index,
            df=df,
            risk=risk,
            lower=lower,
            upper=upper,
            y_adj=y_adj,
            errorbar_kwds=errorbar_kwds,
            ref_kwds=ref_kwds,
            errorbar_color=errorbar_color,
            ref_color=ref_color,
            label=label,
            log_scale=log_scale,
        )

    def draw_horizontal_line(self,
                             y: float,
                             xmin_ax1: float=0, 
                             xmax_ax1: float=1.1, 
                             kwds: dict = None
                             ):
        """Draw horizontal line.
        """
        if kwds is None:
            kwds = dict(lw=1, ls="-", color="black")
        xmin = -1_000_000
        xmax =  1_000_000
        self.ax2.axhline(y=y, xmin=xmin, xmax=xmax, **kwds)
        self.ax1.axhline(y=y, xmin=xmin_ax1, xmax=xmax_ax1, 
                         zorder=-10, clip_on=False, 
                         **kwds)

    def horizontal_variable_separators(self, 
                                       xmin_ax1: float=0, 
                                       xmax_ax1: float=None, 
                                       kwds: dict = None):
        """Draw horizontal lines for seprating variables.

        Args:
            kwds: Passed to ax.axhline function.
        """
        hlines = self.y_index_cate.copy() + 0.5
        xmin, xmax = self.ax2.get_xlim()

        # to cover figure part.
        if xmax_ax1 is None:
            xmax_ax1 = 1 + self.ratio[1]/self.ratio[0]*1.1
        for y in hlines:
            self.draw_horizontal_line(y=y, 
                                      xmin_ax1=xmin_ax1, 
                                      xmax_ax1=xmax_ax1, 
                                      kwds=kwds)

    def embed_strings(self, 
                      col: str, 
                      x: float, 
                      header: str, 
                      fontsize: int = None,
                      y_header: float = 1.0,
                      y_adj : float = 0.0,
                      text_kwds: Optional[dict] = None,
                      header_kwds: Optional[dict] = None,
                      duplicate_hide: bool = False,
                      replace: Optional[dict] = None,
                      df: Optional[pd.DataFrame] = None,
                      ):
        """Embed strings/values of one column with header.

        Args:
            col: Column name for text.
            x: x axis value of text position, ranging from 0 to 1.
            df: Dataframe for another result.
        """
        if df is None:
            df = self.df
        ser = df[col]
        # Drop duplicated items
        if duplicate_hide: 
            cond = ser.duplicated()
            ser = ser.mask(cond, "")

        y_index = self.y_index + y_adj

        vis_utils.embed_strings_forestplot(
            ax=self.ax1,
            ser=ser,
            y_index=y_index,
            x=x,
            header=header,
            fontsize=fontsize,
            y_header=y_header,
            text_kwds=text_kwds,
            header_kwds=header_kwds,
            replace=replace
        )

    def embed_cate_strings(self,
                           col: str,
                           x: float,
                           header: str,
                           fontsize: int = None,
                           y_header: float = 1.0,
                           y_adj : float = 0.0,
                           text_kwds: Optional[dict] = None,
                           header_kwds: Optional[dict] = None,
                           replace: Optional[dict] = None,
                           df: Optional[pd.DataFrame] = None,
                           ):
        """Embed category values on vertically aligned positions.
        The position of strings become different only if self.vertical_align == True.
        """
        if df is None:
            df = self.df
        ser = df[col].drop_duplicates()

        y_index = self.y_index_cate + y_adj

        vis_utils.embed_strings_forestplot(
            ax=self.ax1,
            ser=ser,
            y_index=y_index,
            x=x,
            header=header,
            fontsize=fontsize,
            y_header=y_header,
            text_kwds=text_kwds,
            header_kwds=header_kwds,
            replace=replace
        )






from setuptools import setup
from codecs import open
from os import path

current = path.abspath(path.dirname(__file__))
INSTALL_REQUIRES = [
    'matplotlib>=3.5.1',
]
EXTRAS_REQUIRE = {}

# long_description(後述)に、GitHub用のREADME.mdを指定
with open(path.join(current, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='myforestplot', # パッケージ名(プロジェクト名)
    packages=['myforestplot'], # パッケージ内(プロジェクト内)のパッケージ名をリスト形式で指定
    license='MIT', # ライセンス

    author='toshiakiasakura', # パッケージ作者の名前
    author_email='wordpress.akitoshi@gmail.com', # パッケージ作者の連絡先メールアドレス
    url='https://toshiakiasakura.github.io/myforestplot', # パッケージに関連するサイトのURL(GitHubなど)

    description='Create forestplot by Python.', # パッケージの簡単な説明
    long_description=long_description, # PyPIに'Project description'として表示されるパッケージの説明文
    long_description_content_type='text/markdown', # long_descriptionの形式を'text/plain', 'text/x-rst', 'text/markdown'のいずれかから指定
    keywords='forestplot odds-ratio relative-risk meta-analysis', # PyPIでの検索用キーワードをスペース区切りで指定

    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,

    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.9',
    ], # パッケージ(プロジェクト)の分類。https://pypi.org/classifiers/に掲載されているものを指定可能。
)




from typing import Union, Optional, List, Dict, Tuple, Any, Callable
import pandas as pd
import numpy as np
import statsmodels


def statsmodels_fitting_result_dataframe(
    res,
    alpha: float = 0.05,
    accessor: Callable[[np.array], np.array] = np.exp,
    ) -> pd.DataFrame:
    """Create category and item columns from the statsmodels result.
    Categorical results are divided into original column name (category) and
    its items (item).

    Args:
        res: statsmodels' fitting results.
        alpha: The significance level for the confidence interval.
        accessor: Function to access each model result, which is summarized and displayed.
    """
    df = accessor(res.conf_int(alpha=0.05))
    df["risk"] = accessor(res.params)

    cate = "category"
    df[cate] = np.nan
    rename_dic = {}
    for ind in df.index:
        if "[" in ind:
            s1, s2 = ind.split("[")
            rename_dic[ind] = s2[2:-1]
            # For case of specifying Treatment in formula.
            if "Treatment('" in s1:
                s1 = s1.split(",")[0][2:]
            df.loc[ind, cate] = s1
    df.rename(index=rename_dic, inplace=True)

    # Insert the same name for "category" in case of continuous variables.
    cond = df[cate].isnull()
    df[cate] = df[cate].mask(cond, df.index)
    df = df.reset_index().rename(columns={"index":"item"})

    df.insert(0, column="category", value=df.pop("category"))
    df.insert(1, column="item", value=df.pop("item"))

    # drop Intercept.
    df = df[df["category"] != "Intercept"]

    return df


def add_pretty_risk_column(res: pd.DataFrame, risk: str, lower: str, upper: str,
                           fml: str = ".2f", ref: str = "Ref."
                           ) -> pd.Series:
    """Add prrety risk string column.

    Args:
        res: Dataframe contaning points and confidence intervals.
        risk: point estimates of risk column name.
        lower: lower confidence interval column name.
        upper: upper confidence interval column name.
        fml: formula for f string.
        ref: if point esitmate column is empty, insert this string.
    """
    def f(x):
        risk_v = x[risk]
        lower_v = x[lower]
        upper_v = x[upper]
        s = f"{risk_v:{fml}} ({lower_v:{fml}}, {upper_v:{fml}})"
        return s

    ser = (res.apply(f, axis=1)
           .mask(res[risk].isnull(), ref)
           )
    return ser


def count_category_frequency(df: pd.DataFrame,
                             categorical_cols: List[str],
                             impute_continuous: bool = True,
                             ) -> pd.DataFrame:
    """Count category frequency.

    Args:
        df: Original dataframe.
        categorical_cols: Columns for categorical variables.
        impute_continuous: columns not specified as categorical_cols were
            imputed for item and number of observations (nobs).
    """
    n = df.shape[0]
    sers = [(df[c]
             .value_counts()
             .to_frame()
             .stack()
             )
             for c in categorical_cols]
    ser_sum = pd.concat(sers)
    df_nobs = (ser_sum
               .reset_index()
               .rename(columns={"level_0": "item",
                                "level_1": "category",
                                0:"nobs"}
                       )
               )

    df_nobs.insert(0, column="category", value=df_nobs.pop("category"))
    df_nobs.insert(1, column="item", value=df_nobs.pop("item"))
    return df_nobs


def sort_category_item(df_: pd.DataFrame,
                       order: List[str],
                       item_order: Dict[str, List[str]] = None,
                       ) -> pd.DataFrame:
    """Sort category and item based on categorical values.

    Args:
        df_ : dataframe containing category and item.
        categorical: Dictionary containing column names and its order of items.
        order : if specified, category is ordered based on this variable.

    """
    if item_order is None:
        item_order = {}

    df_sorted = pd.DataFrame()
    for c in order:
        cond = df_["category"] == c
        dfM = df_[cond]
        if c in item_order.keys():
            lis = item_order[c]
            sort_dic = {l:i for i,l in enumerate(lis)}
            dfM = dfM.sort_values(by="item",key=lambda x: x.replace(sort_dic))
        df_sorted = pd.concat((df_sorted, dfM), axis=0)
    return df_sorted


def statsmodels_pretty_result_dataframe(
    data: pd.DataFrame,
    res,
    order: List[str],
    cont_cols: Optional[List[str]] = None,
    item_order: Dict[str, List[str]] = None,
    fml: str = ".2f",
    accessor: Callable[[np.array], np.array] = np.exp,
) -> pd.DataFrame:
    """Obtain pretty result dataframe from statsmodels results.
    Fitting coefficients are converted by np.exp.

    Args:
        data: original dataframe.
        res: statsmodels results.
        categorical: Dictionary containing column names and its order of items.
        order : if specified, category is ordered based on this variable.
        fml: formula for f string of pretty risk.
        accessor: Function to access each model result, which is summarized and displayed.
    """
    if res.nobs != data.shape[0]:
        raise Exception(("Some observations were dropped when fitted, "
                         "check number of observations"
                        ))
    if cont_cols is None:
        cate_cols = order
    else:
        cate_cols = [c for c in order if not c in cont_cols]
    df_res = statsmodels_fitting_result_dataframe(res, alpha=0.05, accessor=accessor)
    df_nobs = count_category_frequency(data, cate_cols)
    df_sum = pd.merge(df_res, df_nobs,
                      on=["category", "item"],
                      validate="1:1",
                      how="outer")
    df_sum = sort_category_item(df_sum, order=order, item_order=item_order)
    df_sum["risk_pretty"] = add_pretty_risk_column(df_sum,
                                                   risk="risk",
                                                   lower=0,
                                                   upper=1,
                                                   fml=".2f"
                                                   )
    return df_sum

